//*****************************************************************//
//                                                                 //
//  TBitList Object                                                //
//  Copyright BrandsPatch LLC                                     //
//  http://www.explainth.at                                        //
//                                                                 //
//  All Rights Reserved                                            //
//                                                                 //
//  Permission is granted to use, modify and redistribute          //
//  the code in this Delphi unit on the condition that this        //
//  notice is retained unchanged.                                  //
//                                                                 //
//  BrandsPatch declines all responsibility for any losses,        //
//  direct or indirect, that may arise  as a result of using       //
//  this code.                                                     //
//                                                                 //
//                                                                 //
//*****************************************************************//
unit BitList;

interface

uses Windows,SysUtils,Classes;

type TByteList = array[0..8191] of Byte;
     PByteList = ^TByteList;

type TBitList = class(TObject)
private
  FBytes:PByteList;
  FSize:Word;
  FPosition:Word;
  FMaxPos:Integer;
  function GetBits(Index:Integer):Boolean;
  procedure SetBits(Index:Integer;Value:Boolean);
  procedure Expand;
public
  constructor CreateEx;
  constructor CreateFromStream(AStream:TStream);
  destructor Destroy;override;
  procedure WriteToStream(AStream:TStream);
  procedure AddBit(Value:Boolean);
  function FetchBit:Boolean;
  procedure ReSet;
  property Bits[Index:Integer]:Boolean read GetBits write SetBits;
end;

type EBitList = class(Exception);

implementation

constructor TBitList.CreateEx;
begin
  inherited Create;
  FSize:=4;FMaxPos:=31;
  FBytes:=AllocMem(FSize);//32 bits @ 8 bits per byte
end;

constructor TBitList.CreateFromStream(AStream:TStream);
begin
  with AStream do
  begin
    ReadBuffer(FSize,sizeof(FSize));
    FBytes:=AllocMem(FSize);
    ReadBuffer(FBytes^,FSize);
  end;
end;

destructor TBitList.Destroy;
begin
  ReAllocMem(FBytes,0);
  inherited;
end;

procedure TBitList.WriteToStream(AStream:TStream);
var ASize:Word;
begin
  ASize:=(FPosition div 8) + ord(FPosition mod 8 > 0);
  with AStream do
  begin
    WriteBuffer(ASize,sizeof(ASize));
    WriteBuffer(FBytes^,ASize);
  end;
end;

function TBitList.GetBits(Index:Integer):Boolean;
var AByte,ABit:Integer;
begin
  if (Index < 0) or (Index > FMaxPos) then raise EBitList.Create('Bad Index');
  AByte:=Index div 8;
  ABit:=Index mod 8;
  Result:=(FBytes^[AByte] and (1 shl ABit) > 0);
end;

procedure TBitList.SetBits(Index:Integer;Value:Boolean);
var AByte,ABit:Integer;
begin
  if (Index < 0) or (Index > FMaxPos) then raise EBitList.Create('Bad Index');
  AByte:=Index div 8;
  ABit:=Index mod 8;
  FBytes^[AByte]:=(FBytes^[AByte] and ($FF xor (1 shl ABit))) or (ord(Value) shl ABit);
end;

procedure TBitList.AddBit(Value:Boolean);
var AByte,ABit:Integer;
begin
  if (FPosition = FMaxPos) then Expand;
  AByte:=FPosition div 8;
  ABit:=FPosition mod 8;
  FBytes^[AByte]:=(FBytes^[AByte] and ($FF xor (1 shl ABit))) or (ord(Value) shl ABit);
  inc(FPosition);
end;

procedure TBitList.Expand;
begin
  inc(FSize,4);//room for another 32 bits
  ReAllocMem(FBytes,FSize);
end;

function TBitList.FetchBit:Boolean;
var AByte,ABit:Integer;
begin
  AByte:=FPosition div 8;
  ABit:=FPosition mod 8;
  Result:=(FBytes^[AByte] and (1 shl ABit) > 0);
  inc(FPosition);
end;

procedure TBitList.ReSet;
begin
  FSize:=4;
  FPosition:=0;
  FMaxPos:=31;
  ReAllocMem(FBytes,4);
  FillChar(FBytes^,4,0);
end;

end.
